#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QDebug>

//PBI_Struct *pbi_data = NULL;

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    //Setup Initial state of GUI objects and connect signal/slots
    ui->setupUi(this);
    ui->progressBar->setVisible(FALSE);
    ui->pushKillBuild->setEnabled(FALSE);
    ui->gui_app->setEnabled(FALSE);
    ui->create_button->setEnabled(FALSE);
    ui->pushBuildPBI->setEnabled(FALSE);
    ui->pushFindIcon->setEnabled(FALSE);
    ui->pushSaveLog->setEnabled(FALSE);
    ui->tabWidget->setCurrentIndex(0); //default to the Modules Tab
    connect(ui->pushResetModule,SIGNAL(clicked()),this,SLOT(Reset_Form()) );
    
    SetupDefaults();
}

MainWindow::~MainWindow()
{
    if (pbi_data)
    {
        Clean_Up(pbi_data, TRUE);
    }
    //delete ui;
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::SetupDefaults(){
  pbi_data = NULL;
  PROGVERSION = "1.0";
  PROG_DIR = QDir::homePath() + "/EasyPBI";
  PBIOUT_DIR= PROG_DIR + "/PBI";
  MODOUT_DIR= PROG_DIR + "/Modules";  
  PBISETTINGS_FILE= PROG_DIR + "/.savedPBISettings.conf";
  PORTS_DIR = "";
  PORT_SELECTED = "";

  //Check for EasyPBI directory structure and create it if it is not there
  if( !QDir(PROG_DIR).exists() ){
    QDir dir;
    dir.root();
    if( !dir.mkdir(PROG_DIR) ){ //Create the program directory
      //Could not create the Program directory
      QMessageBox::warning(this, "Error: Permissions", "Invalid permissions to create EasyPBI directory in home folder");
      return;
    }
  }  
  if( !QDir(MODOUT_DIR).exists() ){
    QDir dir;
    dir.root();
    if( !dir.mkdir(MODOUT_DIR) ){ //Create the program directory
      //Could not create the Program directory
      QMessageBox::warning(this, "Error: Permissions", "Invalid permissions to create EasyPBI directory in home folder");
      return;
    }
  }  
  if( !QDir(PBIOUT_DIR).exists() ){
    QDir dir;
    dir.root();
    if( !dir.mkdir(PBIOUT_DIR) ){ //Create the program directory
      //Could not create the Program directory
      QMessageBox::warning(this, "Error: Permissions", "Invalid permissions to create EasyPBI directory in home folder");
      return;
    }
  }  
  //Check for the ports Directory and Set the internal variable 'PORTS_DIR'
  if(!Check_For_Ports(true)){ return; }
  //Load the defaults for the PBI build settings
  if( QFile(PBISETTINGS_FILE).exists() ){
    qDebug() << "PBI Settings file detected at:"<< PBISETTINGS_FILE;
    loadPBISettings();
  }else{
    ui->linePBIOutDir->setText(PBIOUT_DIR);
  }
  
}

void MainWindow::Reset_Form()
{
    if (pbi_data)
    {
      Clean_Up(pbi_data, TRUE);
      pbi_data = NULL;
    }
    ui->port_name->setText("");
    ui->port_category->setText("");
    ui->author->setText("");
    ui->icon_path->setText("");
    ui->gui_app->setChecked(true);
    ui->website->setText("");
    ui->status_bar->setText("Please push 'New Module' to get started.");
    ui->port_selected->setText("");
    //Reset GUI buttons
    ui->gui_app->setEnabled(FALSE);
    ui->create_button->setEnabled(FALSE);
    ui->pushFindIcon->setEnabled(FALSE);
}

void MainWindow::on_pushFindIcon_clicked()
{
    QString icon_name;
    icon_name = QFileDialog::getOpenFileName(this, tr("Select Icon"), "~", tr("Image Files (*.png)"));
    if( !icon_name.isEmpty() ){
      ui->icon_path->setText(icon_name);
    }
}

void MainWindow::on_actionAbout_EasyPBI_triggered()
{
    QMessageBox::about(this, "About EasyPBI", "EasyPBI "+PROGVERSION+"\nCreate PBI modules quickly and easily.\nhttp://makeapbi.sourceforge.net\nWritten by Jesse Smith and Ken Moore");
}

void MainWindow::on_actionAppCafe_Integration_triggered(){
  QString instructions = "Steps to integrate your PBI modules with the official PC-BSD Repository:\n\n  1) Ensure the module works by building and testing the PBI with EasyPBI, making adjustments to the module as needed.\n  2) Check that the module does not already exist in the PC-BSD repository at: http://trac.pcbsd.org/browser/pbi/modules\n  3) E-mail the compressed module file to: pbi-dev@lists.pcbsd.org\nThe module will require re-compression (tar czf 'new-filename.tar.gz' 'module-directory') if modifications were made to the module.\n\nTo place the module into an alternate PBI repository, please contact the maintainer of that repository for instructions.";
  QMessageBox::about(this, "EasyPBI: AppCafe Integration", instructions);
}

void MainWindow::on_pushNewModule_clicked()
{
    QString module_path;
    int status;
    char *argv[] = {"-9", "-s", ""};

    // clear out any existing module
    if (pbi_data)
    {
        Reset_Form();
    }

    // check to see if the ports tree exists
    if (!Check_For_Ports(true)){return;}

    // get new port to work on
    module_path = QFileDialog::getExistingDirectory(this, "Select Port", PORTS_DIR);
    if(module_path.isEmpty()){return;} //action cancelled or closed
    PORT_SELECTED = module_path.section("ports/",1,1);
    pbi_data = Init_Data();
    if (! pbi_data)
    {
        QMessageBox::about(this, "Error", "Could not set up memory for this module.");
	return;
    }
    // get the name, arguments
    argv[2] = strdup(qPrintable(module_path));
    Get_Options(pbi_data, 3, argv);
    // confirm port exists
    status = Port_Exists(pbi_data);
    if (! status)
    {
        QMessageBox::about(this, "Error", "This port does not exist.");
        Clean_Up(pbi_data, TRUE);
        return;
    }
    Get_Name(pbi_data);
    Get_Icon(pbi_data);
    Get_Category(pbi_data);
    Get_Website(pbi_data);
    Get_Author(pbi_data);
    ui->author->setText(pbi_data->author_name);
    ui->port_category->setText(pbi_data->category);
    ui->port_selected->setText(PORT_SELECTED);
    if (pbi_data->icon_name)
        ui->icon_path->setText(pbi_data->icon_name);
    ui->port_name->setText(pbi_data->port_name);
    ui->website->setText(pbi_data->web_site);
    ui->status_bar->setText("Please fill out the above fields and click Create Module.");
    //Enable the GUI Buttons
    ui->gui_app->setEnabled(TRUE);
    ui->create_button->setEnabled(TRUE);
    ui->pushFindIcon->setEnabled(TRUE);
    status = Is_GUI(pbi_data);
    if (status)
        ui->gui_app->setChecked(true);
    else
        ui->gui_app->setChecked(false);
}

void MainWindow::on_create_button_clicked()
{
    char *home_dir;
    char final_output[1024];
    BIN_FILE *current_exe;
    //int status;
    FILE *my_file;

    if (! pbi_data)
    {
        QMessageBox::about(this, "Error", "Please push 'New Module' first.");
        return;
    }
    // find where we want to save the module
    ui->status_bar->setText("Please wait while we build the module...");
    // save text fields to data structure
    if (pbi_data->author_name) free(pbi_data->author_name);
    if (pbi_data->category) free(pbi_data->category);
    if (pbi_data->icon_name) free(pbi_data->icon_name);
    if (pbi_data->web_site) free(pbi_data->web_site);
    pbi_data->author_name = strdup( qPrintable(ui->author->text()));
    pbi_data->category = strdup( qPrintable(ui->port_category->text()));
    pbi_data->icon_name = strdup( qPrintable(ui->icon_path->text()));
    pbi_data->web_site = strdup( qPrintable(ui->website->text()));
    if( ui->gui_app->isChecked() ){pbi_data->console_app = FALSE;}
    else{pbi_data->console_app = TRUE;}
    // chdir to save location
    home_dir = getenv("HOME");
    chdir(home_dir);
    chdir("EasyPBI/Modules"); //Already exists (checked before now)
    //mkdir("Modules", 0777);   

    // build module
    Create_Module_Template9(pbi_data);
    Get_Main_Exe_Smart(pbi_data);
    Get_More_Exe(pbi_data);
    Copy_Icon9(pbi_data);
    Write_Conf9(pbi_data);
    Write_Menu(pbi_data, NULL);
    current_exe = pbi_data->more_exe;
    while (current_exe)
    {
        Write_Menu(pbi_data, current_exe->path);
        current_exe = (BIN_FILE *) current_exe->next;
    }

    // create tarball of module
    sprintf(final_output, "tar czf %s.tar.gz %s", pbi_data->port_name9, pbi_data->port_name9);
    system(final_output);
    sprintf(final_output, "%s.tar.gz", pbi_data->port_name9);
    my_file = fopen(final_output, "r");
    if (my_file)
    {  
        sprintf(final_output, "Done. Your module is located at %s/EasyPBI/Modules/\nPlease view the 'AppCafe Integration' page for further instructions", home_dir);
        fclose(my_file);
    }
    else
        sprintf(final_output, "Something went wrong creating the module. There may be a permission error creating new directories in your home.");
    ui->status_bar->setText(final_output);
}

void MainWindow::on_port_info_button_clicked()
{
   //Check if a port is selected
   if(! pbi_data){
     QMessageBox::warning(this, "Error: Port Information", "Please select a port first");
     return;
   }
   //Open up the freshports.org page for the selected port
   QString target_url = "http://freshports.org/" + PORT_SELECTED;
   QDesktopServices::openUrl(QUrl(target_url));
}


int MainWindow::Check_For_Ports(bool show_warning)
{
    QStringList portsLocations;
    //Set the locations to search for the ports tree
    portsLocations << "/usr/ports" << PROG_DIR +"/ports";
    //Search the locations
    for(int i=0; i<portsLocations.size(); i++){
      if( QDir(portsLocations[i]).exists() ){
	PORTS_DIR=portsLocations[i];
	return TRUE;
      }
    }
    //No ports tree detected - Display warning
    if (show_warning)
        QMessageBox::warning(this, "Ports Not Installed","The FreeBSD ports tree is not installed for this system or user.\nPlease Install the ports tree as root to make it available for all users, or push 'File'->'Get Ports' to retrieve the ports tree for this user ONLY");
    return FALSE;
}

void MainWindow::on_actionGet_Ports_triggered()
{
    //Check to see if the ports are already installed
    if(Check_For_Ports(false)){
      if(PORTS_DIR=="/usr/ports"){
	QMessageBox::warning(this, "Ports Installed","The FreeBSD ports tree is already available on this system");
	return;
      }else{
	QMessageBox verify;
	verify.setText("The FreeBSD ports tree is already available for this user");
	verify.setInformativeText("Do you wish to update the user ports tree? (requires re-download)");
	verify.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
	verify.setDefaultButton(QMessageBox::Yes);
	int ret = verify.exec();
	if(ret != QMessageBox::Yes){return;}
      }
    }
    //Change to the 'Modules' Tab for download status
    ui->tabWidget->setCurrentIndex(0);
    
    //Start the Download
    QNetworkReply *currentDL;
    QNetworkAccessManager *manager = new QNetworkAccessManager(this);
    connect(manager, SIGNAL(finished(QNetworkReply*)), this, SLOT(downloadFinished(QNetworkReply*)) );
    currentDL = manager->get(QNetworkRequest(QUrl("ftp://ftp.freebsd.org/pub/FreeBSD/ports/ports/ports.tar.gz")));
    
    ui->status_bar->setText("Downloading the FreeBSD ports tree: Please Wait.");
    
    //Make the Progress Bar visible and start updating
    ui->progressBar->setValue(0);
    ui->progressBar->setVisible(TRUE);
    connect(currentDL, SIGNAL(downloadProgress(qint64,qint64)), this, SLOT(updateProgress(qint64,qint64)) );
}

void MainWindow::downloadFinished(QNetworkReply *reply){
  //Disable the progress bar
  ui->progressBar->setVisible(FALSE);
  //See if the Download was successful
  qDebug() << "Ports download completion code:" << reply->error();
  if( reply->error() != QNetworkReply::NoError){
    QMessageBox::warning(this, "Download Failed","Downloading the FreeBSD ports tree failed. Please try again.");
    return;
  }
  //Download Succesful, save it to file
  QFile file(PROG_DIR+"/ports.tar.gz");
  if( !file.open(QIODevice::WriteOnly)){   //if file could not be opened
    QMessageBox::warning(this, "Saving Failed","Saving the FreeBSD ports tree file failed. Please try again.");
    reply->deleteLater();
    return;
  }
  file.write(reply->readAll());
  file.close();
  //Saving the file successful, delete the downloaded data from memory
  reply->deleteLater();
  ui->status_bar->setText("Download finished. Extracting the ports tree into the EasyPBI directory.");
  QTimer::singleShot(10,this, SLOT(extractPT()) );
  return;
}

void MainWindow::extractPT(){
  //See if the directory already exists, and remove it if it does
  if(QDir(PROG_DIR+"/ports").exists()){
    QString tmp = "cd "+PROG_DIR+"; rm -rf ports/";
    system(tmp.toLatin1());
  }
  //Unpack the file
  QString cmd = "cd " + PROG_DIR + "; tar xvf ports.tar.gz";
  system(cmd.toLatin1());
  
  //Delete the file
  QFile file(PROG_DIR+"/ports.tar.gz");  
  file.remove();
  //Let the user know that the ports tree was successfully added
  ui->status_bar->setText("The FreeBSD ports tree has been successfully added to the EasyPBI directory.\nPlease select 'New Module' from the file menu to continue.");
  return;
}

void MainWindow::updateProgress(qint64 bytesReceived, qint64 bytesTotal){
  //Determine the percentage
  int percent = (bytesReceived*100)/bytesTotal;
  //Update the progress bar
  ui->progressBar->setValue(percent);
  ui->progressBar->update();
}

void MainWindow::on_pushChangeOutDir_clicked(){
  QString outDir_path; 
  outDir_path = QFileDialog::getExistingDirectory(this,"Select PBI Output Directory",PROG_DIR+"/PBI");
  if(outDir_path.isEmpty()){return;} //action cancelled or closed
  PBIOUT_DIR = outDir_path;
  ui->linePBIOutDir->setText(outDir_path);
}

void MainWindow::on_pushChangeSigFile_clicked(){
  QString outDir_path;
  outDir_path = QFileDialog::getOpenFileName(this, "Select Digital Signiture File", PROG_DIR,"Signature Files (*)");
  if(outDir_path.isEmpty()){return;} //action cancelled or closed
  PBIOUT_DIR = outDir_path;
  ui->linePBIDigSigFile->setText(outDir_path);
}

void MainWindow::on_pushSavePBISettings_clicked(){
  QFile *file = new QFile(PBISETTINGS_FILE);
  //Open the File
  file->open(QIODevice::WriteOnly);
  QTextStream out(file);
  //Save the default lines
  out << "PBI default build settings: ---DELETE TO RESET DEFAULTS, DO NOT MODIFY--\n";
  out << PROGVERSION+"\n";
  out << ui->linePBIOutDir->text() + "\n";
  out << ui->linePBIDigSigFile->text() + "\n";
  if(ui->checkTMPFS->isChecked()){out << 1;}
  else{out << 0;}
  
  file->close();
  //Inform the user that the save is complete
  QMessageBox::information(this, "Settings Saved","PBI build settings successfully saved as the user defaults.");
}

void MainWindow::loadPBISettings(){
  qDebug() << "Loading Previously Saved PBI Settings";
  QFile *file = new QFile(PBISETTINGS_FILE);
  //Open the File
  file->open(QIODevice::ReadOnly);
  QTextStream in(file);
  //Load the lines into the GUI
  in.readLine(); //skip the label line
  QString ver = in.readLine(); //version of the program when it saved the settings
  ui->linePBIOutDir->setText(in.readLine()); //output directory
  ui->linePBIDigSigFile->setText(in.readLine()); //Digital Signiture File
  QString line = in.readLine(); //TMPFS checked
  int chked = line.toInt();
  if(chked==1){ui->checkTMPFS->setChecked(TRUE);}
  else{ui->checkTMPFS->setChecked(FALSE);}

  file->close();
}

void MainWindow::on_pushSelectModule_clicked(){
  QString outDir_path;
  outDir_path = QFileDialog::getExistingDirectory(this, "Select Module to Build", PROG_DIR+"/Modules");
  if(outDir_path.isEmpty()){return;} //action cancelled or closed
  ui->linePBIModuleSelected->setText(outDir_path);
  ui->pushBuildPBI->setEnabled(TRUE);
}

void MainWindow::on_pushBuildPBI_clicked(){

  //Check for the Port tree
  if(!Check_For_Ports(true)){return;}
  
  //Check GUI to make sure settings are set before running
  bool gostatus = TRUE;
  QString outdir = ui->linePBIOutDir->text();
  if (outdir.isEmpty() ){gostatus=FALSE;}
  QString sigfile = ui->linePBIDigSigFile->text(); //this one can be empty
  QString modDir = ui->linePBIModuleSelected->text();
  if(modDir.isEmpty()){gostatus=FALSE;}
  if(!gostatus){
    QMessageBox::warning(this,"Error","Invalid PBI Settings");
    return;
  }
  
  //Receive User verification before beginning build process due to:
  //  -- long time required, internet connection required, root permissions required
  QMessageBox verify;
  verify.setText("Are you sure you wish to start the PBI build?");
  verify.setInformativeText("This requires an active internet connection and root permissions. The time required to build a PBI varies depending upon system specifications and ports to be compiled.");
  verify.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
  verify.setDefaultButton(QMessageBox::Yes);
  int ret = verify.exec();
  if(ret != QMessageBox::Yes){return;}
  
  //Setup the pbi_makeport command from GUI settings
  QString cmd = "gksu -a \"pbi_makeport";
  cmd += " -c " + modDir;
  if(PORTS_DIR != "/usr/ports"){ cmd += " -d " + PORTS_DIR; }
  cmd += " -o " + outdir;
  cmd += " --delbuild";
  if( ui->checkTMPFS->isChecked() ){ cmd += " --tmpfs"; }
  if(!sigfile.isEmpty()){ cmd += " --sign " + sigfile; }
  cmd += "\"";
  qDebug() << "Build PBI command created:"<<cmd;
  
  //Setup the displays
  ui->pushKillBuild->setEnabled(TRUE);
  ui->pushSaveLog->setEnabled(FALSE);
  ui->pushBuildPBI->setEnabled(FALSE); //disable the button so they do not start more than 1 build at a time
  ui->buildOutputArea->clear(); //clear the display in case this is not the first run
  
  //Setup Process connections
  p = new QProcess(this);
  p->setProcessChannelMode(QProcess::MergedChannels);
  connect(p,SIGNAL(readyReadStandardOutput()),this,SLOT(updatePBIBuild()) );
  connect(ui->pushKillBuild,SIGNAL(clicked()),this,SLOT(killPBIBuild()) );
  connect(p,SIGNAL(finished(int,QProcess::ExitStatus)),this,SLOT(PBIbuildFinished(int,QProcess::ExitStatus)) );
  connect(p,SIGNAL(error(QProcess::ProcessError)),this,SLOT(buildError(QProcess::ProcessError)) );
  
  //Start the Process
  qDebug() << "Starting the PBI build process...";
  p->start(cmd);  

}

void MainWindow::updatePBIBuild(){
  QString tmp = p->readAllStandardOutput();
  if( tmp.startsWith("\n") ){tmp.remove(0,0);} //remove newline at the beginning (too much whitespace in log)
  if( tmp.endsWith("\n") ){tmp.chop(1);} //remove newline at the end (already accounted for by appending)
  if(!tmp.isEmpty()){ ui->buildOutputArea->append( tmp ); }
  //qDebug() << "Update output: " << tmp;
}

void MainWindow::PBIbuildFinished(int exitCode,QProcess::ExitStatus exitStatus){
  //Check to see if the PBI process finished successfully
  qDebug() << "PBI build process Finished" << exitStatus << exitCode;
  if(exitCode == 0){
    if(exitStatus==QProcess::NormalExit){
      QMessageBox::information(this,"PBI Build Success","The PBI finished building successfully");
    }else{
      //The user killed the process - No Message
    }
  }else{
    QMessageBox::warning(this,"PBI Build Failure","The PBI failed to build.\nPlease check the build log to find the cause of the failure and adjust the module accordingly");
  }
  ui->pushBuildPBI->setEnabled(TRUE);
  ui->pushSaveLog->setEnabled(TRUE);
  ui->pushKillBuild->setEnabled(FALSE);
  p->close();
}

void MainWindow::killPBIBuild(){
  if( p->state()==QProcess::Running ){
    //qDebug() << "Process PID:" << p->pid(); 
    p->terminate();
    ui -> buildOutputArea->append("---PBI Build Terminated---");
  }
}

void MainWindow::buildError(QProcess::ProcessError error){
  qDebug() << "QProcess PBI Build error#" << error;
}

void MainWindow::on_pushSaveLog_clicked(){
  //Have user select a filename/location to save the log
  QString filename = QFileDialog::getSaveFileName(this,"Save Log",PROG_DIR+"/build.log","Log Files (*.log *.txt)");
  if(filename.isEmpty()){return;} //User cancelled the process
  //Open the file (as new)
  QFile *file = new QFile(filename);
  file->open(QIODevice::WriteOnly);
  //Save the log into the file
  file->write(ui->buildOutputArea->toPlainText().toUtf8());
  file->close();
}

